/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.0b (Yorick, Vivriel, Thuzzle, Ilie 1992-95)"
#define SAVEFILEVER 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "edit.h"
#include "list.h"
#include "beam.h"

int select();
int pause();
char *strdup();
int strncasecmp();
int setenv();
void unsetenv();

extern int errno, sys_nerr;
extern char *sys_errlist[];


/* 	function declarations		*/
void initcancan();
void read_settings();
void def_error();
void mainlupe();
void printmotd();
void getuserinput();
void show_footnote();
void cleanup();
void process_sub_input();
void process_input();
int printstrlen();
void mark_and_print();
void wrap_out();
void intrsig();
void termsig();
void pipesig();
void denewbiefy();
void get_remote_output();
void writetofile();

/*	GLOBALS		*/

char hostname[LINELEN];
char portnumber[LINELEN];
char initstr[LINELEN];	/* initial string to send on connect */

FILE *capfile;		/* capture file or NULL */

aliasnode *aliases;	/* head of alias list */
actionnode *actions;	/* head of action list */
marknode *markers;	/* head of mark list */
keynode *keydefs;	/* head of key binding list */

#ifndef OMIT_SESSIONS
struct consess {
  int active;
  char *id;
} con_sess[MAXFDSCAN];	/* connections indexed by socket descriptors */
#endif

int sockfd;		/* current socket file descriptor */
fd_set fdset;		/* set of descriptors to select() on */

int waiting_for_pid;	/* = pid, when executing #! or editor */
int wait_result;	/* exit status of the process we waited for */

int linemode;		/* line mode flags (LM_* in cancan.h) */

/* booleans */
int mp_active;		/* 1 = we have #identified ourselves */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* 1 if command is user-entered */
int wrap = 1;		/* 1 = word wrap active */
int speedwalk;		/* speedwalk on? */
int verbatim;		/* 1 = don't expand aliases or process semicolons */
int need_cleanup;	/* need either flashback or to clear a footnote */
int flashback;		/* cursor is on excursion and should be put back */
int edattr;		/* 1 = input line attributes are in effect */

/* for line editing */
int cols_1, lines_1;	/* for efficiency's sake, cols - 1, lines - 1 */
int cols, lines;	/* screen size */
int col0;		/* input line offset (= length of prompt) */
int line0;		/* screen line where the input line starts */
int edlen;		/* length of current input line */
int pos;		/* cursor position in line */
char edbuf[LINELEN]; 	/* line editing buffer */
char prefixstr[LINELEN];/* inserted in the editing buffer each time */
char inserted_next[LINELEN];	/* inserted in buffer just once */

char footnote[LINELEN];	/* displayed at the bottom (just once) */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */
char edattrbeg[CAPLEN];	/* starting input line attributes */
char edattrend[CAPLEN]; /* ending input line attributes */
int excursion;		/* where the excursion is */

/* history buffer */
char *hist[MAXHIST];	/* saved history lines */
int curline;		/* current history line */
int pickline;		/* line to pick history from */

/* completion list */
char *words[WORDSKEPT];
int wordindex;

int main(argc, argv)
int argc; char *argv[];
{
    if(argc < 2 || argc > 3) {
	fprintf(stderr,
		"usage:\t%s host port\nor\t%s definition-file\n",
		*argv, *argv);
	exit(1);
    }
  
#ifdef MOTDFILE
    printmotd();
#endif
    if(argc == 2)	/* one arg, assuming it is definition file name */
	initcancan(argv[1]);
    else {	/* two args: assume they are hostname and port number */
	initcancan(0);
	strcpy(hostname, argv[1]);
	strcpy(portnumber, argv[2]);
	default_keybindings();
    }
#ifndef OMIT_PRINTVER
    printver();
#endif
    mainlupe();
    
    return 1;
}

/*
 * initialize cancan
 */
void initcancan(df)
char *df;
{
    char buf[LINELEN];
    char *p, *deff = df;
    initstr[0] = 0;
    if(df) {
	/*
	 * rules: if cancan dir is set it is searched first.
	 *        if file doesn't exist, it is created there.
	 * If a slash appears in the name, the cancan dir isn't used.
	 */
	if(!strchr(df, '/') && (p = getenv("CANCANDIR"))) {
	    strcpy(buf, p);
	    if(buf[0] && buf[strlen(buf) - 1] != '/')
		strcat(buf, "/");
	    strcat(buf, df);
	    if(access(buf, R_OK | W_OK) == 0 || access(df, R_OK | W_OK) != 0)
		deff = buf;
	}
	setenv("SAVEFILE", deff, 1);
	if(access(deff, R_OK | W_OK) != 0) {
	    printf("Creating %s\nHost name  :", deff);  fflush(stdout);
	    fgets(hostname, LINELEN, stdin);  strtok(hostname, "\n");
	    printf("Port number:");  fflush(stdout);
	    fgets(portnumber, LINELEN, stdin);  strtok(portnumber, "\n");
	    default_keybindings();
	    save_settings();
	} else
	    read_settings(deff);
    }
    if((p = getenv("CANCANEDITOR"))) {
	setenv("EDITOR", p, 1);
	unsetenv("CANCANEDITOR");
    }
    if((p = getenv("CANCANPAGER"))) {
	setenv("PAGER", p, 1);
	unsetenv("CANCANPAGER");
    }
    init_keybindings();
}

/*
 * show current version
 */
void printver()
{
    printf("cancan ver %s%s (%s %s)\n", VERSION,
#if __STDC__
	   "\nCompiled " __DATE__,
#else
	   "",
#endif
#ifdef VT100
	   "vt100-only",
#else
	   "termcaps",
#endif
#ifdef USETERMIO
	   "termio"
#else
	   "BSD sgtty"
#endif
	   );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, LINELEN, f))
	    fputs(line, stdout);
    }
/*    fflush(stdout);*/
}
#endif

/*
 * main loop.
 */
void mainlupe()
{
    fd_set readfds;
    int err;
    struct timeval timeoutbuf, *timeout;
    
    settermcap();
    set_terminal();
    setsigs();
    
    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    connect_sess(NULL, (*initstr ? initstr : NULL));
#else
    if ((sockfd = doconnect(hostname, atoi(portnumber))) < 0)
	syserr("host not responding");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
    if (*initstr)
	parse(initstr, 0, NULL);	/* XXX */
#endif
    
    printf("Type #help for help.\n");
    
    /* make sure the input line is always the bottom one */
    winchsig();
    gotoxy(0, line0 = lines_1);
    
    /*		main lupe	*/
    
    for(;;) {
	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	readfds = fdset;
	if(flashback) {
	    timeoutbuf.tv_sec = 0;
	    timeoutbuf.tv_usec = FLASHDELAY;
	    timeout = &timeoutbuf;
	} else
	    timeout = (struct timeval *)NULL;

	err = select(MAXFDSCAN, &readfds, NULL, NULL, timeout);
	if(need_cleanup) cleanup();
	if(err < 0) {
	    if(errno != EINTR) syserr("select");
	} else {
#ifndef OMIT_SESSIONS
	    if (FD_ISSET(0, &readfds)) {
		err--;
		sockfd = main_sess();
		getuserinput();
	    }
	    for (sockfd = 3; err && (sockfd < MAXFDSCAN); sockfd++)
		if (FD_ISSET(sockfd, &readfds)) {
		    err--;
		    get_remote_output();
		}
#else
	    if (FD_ISSET(0, &readfds))
		getuserinput();
	    if (FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
	if(*footnote) show_footnote();
    }
}

/*
 * get output from the socket and process/display it.
 */
void get_remote_output()
{
  static char buffer[BUFSIZE];
  int got = read_socket(buffer, sizeof(buffer) - 1);
  if (!got) return;		/* leave room  ^^^ for terminating zero */

#ifdef DEBUGCODE
  /* debug code to see in detail what strange codes come from the server */
   {
     int i; unsigned char c;
     printf("{");
     for(i = 0; i < got; i++) 
      {
	c = buffer[i];
	printf(isprint(c) ? "%c" : "[%d]", c);
      }
     printf("}\n");
   }
#endif
  
#ifndef OMIT_SESSIONS
  if (con_sess[sockfd].active)		/* process only active sessions */
#endif
   {
     if(linemode & LM_CHAR) 
      {
	/* char-by-char mode: just display output, no fuss */
	write(1, buffer, got);
      } 
     else 
      {
	/* line-at-a-time mode: process input in a number of ways */
	clearinpline();

	if(capfile) writetofile(capfile, buffer, got);
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
#ifndef OMIT_SESSIONS
	if(!is_main_sess()) process_sub_input(buffer, got);
	else
#endif
	process_input(buffer, got);

	redraw_input();
      }
   }
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * assert(sizeof(buf) >= size + 1)
 */
void process_sub_input(buf, size)
char *buf; int size;
{
    char *line, *p;
    int i;

    if(col0) printf("%s%s", begoln, clreoln);
    for(i = 0; i < size; i++) {
	line = buf + i;
	for(p = line; i < size && buf[i] != '\n'; i++)
	    if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	*p = '\0';		/* there is room */
	/* FEATURE: no wrapping, no marking in subsidiary session */
	if(search_action(line)) {
	    printf("##%s> %.*s\n", con_sess[sockfd].id,
		    cols_1 - 4 - (int)strlen(con_sess[sockfd].id), line);
	    if(line0 < lines_1) line0++;
	}
    }
    check_prompt();
}
#endif

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * assert(sizeof(buf) >= size + 1)
 */
void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess = 0;
    char *line, *p;
    int i;

    if(data_sent) {
	data_sent = 0;
	if(esess) esess->mode = 'C';	/* sending cancels the session */
    }

    /* process block one line at a time: */
    for(i = 0; i < size; ) {
	line = buf + i;
	if(esess || (mp_active && !strncmp(MPI, line, strlen(MPI))))
	    i += mp_receive(&esess, line, size - i);
	else {
	    for(p = line; i < size && buf[i] != '\n'; i++)
		if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	    *p = '\0';		/* there is room */
	    if(i == size && buf - line + i + col0 < cols) {
		/* This is probably a prompt: strip CR */
		/* FEATURE: '\r', '\b' ... */
		if(*line == '\r') {
		    while(*++line == '\r');
		    strcpy(promptstr, line);
		} else if(strchr(line, '\b')) {
		    strcpy(promptstr, line);
		} else
		    strcat(promptstr, line);
		col0 = printstrlen(promptstr);
		mark_and_print(line);
		search_action(line);
		return;
	    }
	    i++;
	    if(search_action(line)) {		/* only if no #gag */
		/* FEATURE: if line contains control chars or escape sequence,
		  it ought to be printed without wrapping? */
		if(wrap)
		    wrap_out(line);
		else {
		    mark_and_print(line);
		    printf("\n");
		    if(line0 < lines_1) line0++;
		}
		promptstr[0] = col0 = 0;
	    }
	}
    } /* end for */
}

/*
 * Determine the printed length of a string. This is sometimes less than
 * the string length since it might contain escape sequences. Treated as such
 * are "<esc> [ <non-letters> <letter>", "<esc> <non-[>", "<control-char>".
 * This is not entirely universal but covers the most common cases (i.e. ansi)
 */
int printstrlen(s)
char *s;
{
    int l;
    enum { NORMAL, ESCAPE, BRACKET } state = NORMAL;
    for(l = 0; *s; s++) {
	switch(state) {
	case NORMAL:
	    if(*s == '\033') {
		state = ESCAPE;
	    } else if(*s >= ' ') {
		l++;
	    }
	    break;

	case ESCAPE:
	    state = (*s == '[') ? BRACKET : NORMAL;
	    break;

	case BRACKET:
	    if(isalpha(*s))
		state = NORMAL;
	    break;
	}
    }
    return l;
}

/*
 * print line, scanning for marked text
 * (Needs to be done: match marked text only at word boundaries)
 */
void mark_and_print(line)
char *line;
{
    marknode *m, *np;
    char begin[CAPLEN], end[CAPLEN];

    for(m = 0, np = markers; np; np = np->next)
	if((np->pos = strstr(line, np->pattern)))
	    if(!m || m->pos > np->pos) m = np;
    while(m) {
	attr_string(m->attrcode, begin, end);
	printf("%.*s%s%s%s", m->pos - line, line, begin, m->pattern, end);
	line = m->pos + strlen(m->pattern);
	if(!*line) return;
	for(m = 0, np = markers; np; np = np->next)
	    if(np->pos)
		if(np->pos >= line || (np->pos = strstr(line, np->pattern)))
		    if(!m || m->pos > np->pos) m = np;
    }
    fputs(line, stdout);
}

/*
 * write string to stdout, wrapping words to the next line.
 */
void wrap_out(s)
char *s;
{
    char buf[LINELEN];
    int i, k, width;
    if(*s == '\r') {
	while(*++s == '\r');
	if(*s) printf("\r");
    }
    if(!*s)
	printf("\n");
    else for(width = cols_1 - col0; ; width = cols_1) {
	width += wrapglitch;
	for(k = i = 0; s[i] && ++i < width; )
	    if(s[i] == ' ') k = i;
	if((!k || !s[i]) && !(k = i)) break;
	strncpy(buf, s, k);  buf[k] = 0;
	mark_and_print(buf);
	printf("\n");
	if(line0 < lines_1) line0++;
	for(s += k; *s == ' '; s++);
    }
}

/*
 * write string to file, stripping \r
 */
void writetofile(f, str, len)
FILE *f; char *str; int len;
{
    for(; len--; str++) if(*str != '\r') putc(*str, f);
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
void getuserinput()
{
    int i;
    unsigned char c;
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static nchars = 0;		/* number of them */
    keynode *p;

    /* We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line more, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    while((i = read(0, (char*)&c, 1)) < 0 && errno == EINTR);
    if(i != 1)
	syserr("read from stdin");
    if(linemode & (LM_CHAR | LM_NOECHO)) {
	if(linemode & LM_CHAR) {
	    if(write(sockfd, (char*)&c, 1) != 1)
		syserr("write single char to socket");
	    if(!(linemode & LM_NOECHO)) {
		putchar((int)c);
	    }
	} else {
	    /* sending password (line mode, no echo) */
	    edbuf[edlen++] = c;
	    if(c == '\n' || edlen >= LINELEN - 2) {
		printf("\n");
		send_to_host(edbuf, edlen);
		edlen = 0;
	    }
	    edbuf[pos = edlen] = 0;
	}
    } else {
	/* normal mode (echo, line buffered) */
	if(metakey && (c & 0x80)) { /* did meta key set the high bit? */
	    typed[nchars++] = '\033'; /* prepend esc */
	    c &= 0x7f;
	}
	/* search function key strings for match */
	typed[nchars++] = c;
	for(p = keydefs;
	    p && strncmp(typed, p->sequence, nchars);
	    p = p->next);
	if(!p) {
	    insert_char((int)c);
	    nchars = 0;
	} else if(!p->sequence[nchars]) {
	    p->funct(p->call_data);
	    nchars = 0;
	}
    }
}


/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
void flashparen(c)
int c;
{
    int lev, i;
    for(i = pos - 1, lev = 0; i >= 0; i--)
	if(ISRPAREN(edbuf[i])) {
	    lev++;
	} else if(ISLPAREN(edbuf[i])) {
	    lev--;
	    if(lev == 0)
		if(LPAREN(c) == edbuf[i])
		    break;
		else
		    i = -1;
	}
    if(i >= 0) {
	optmove(CURLINE(pos), CURCOL(pos), CURLINE(i), CURCOL(i));
	need_cleanup = flashback = 1;
	excursion = i;
    }
}

/*
 * display a one-line footnote at the bottom of the screen
 */
void show_footnote()
{
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    if(edattr) { edattr = 0; fputs(edattrend, stdout); }
    optmove(to_l, to_c, lines_1, 0);
    if(CURLINE(edlen) >= lines_1) {
	line0--;  to_l--;  printf("\n");
    }
    fputs(footnote, stdout);
    gotoxy(to_c, to_l);
    need_cleanup = 1;
}

/*
 * clear the footnote or put cursor back where it belongs
 */
void cleanup()
{
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    need_cleanup = 0;
    if(flashback) {
	flashback = 0;
	optmove(CURLINE(excursion), CURCOL(excursion), to_l, to_c);
    } else {
	*footnote = 0;
	optmove(to_l, to_c, lines_1, 0);
	fputs(clreoln, stdout);
	optmove(lines_1, 0, to_l, to_c);
    }
}

/*
 * swap editline with a history line
 */
void swap_history(line)
int line;
{
    if(hist[line]) {
	if(hist[pickline]) free(hist[pickline]);
	hist[pickline] = strdup(edbuf);
	pickline = line;
	clearinpline();
	strcpy(edbuf, hist[line]);
	pos = edlen = strlen(edbuf);
	redraw_input();
    }
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
void put_in_history(str)
char *str;
{
    char *p;
    if(hist[curline]) free(hist[curline]);
    hist[curline] = strdup(str);
    if(++curline == MAXHIST) curline = 0;
  
    /* split into words and put into completion list */
    for(p = strtok(str, DELIM); p; p = strtok(NULL, DELIM)) {
	if(strlen(p) >= 3 && p[0] != '#')	/* no commands/short words */
	    put_word(p);
    }
}

/*
 * put word in word completion list
 */
void put_word(word)
char *word;
{
  if(words[wordindex]) free(words[wordindex]);
  words[wordindex] = strdup(word);
  if(++wordindex == WORDSKEPT) wordindex = 0;
}

/*
 * match and complete a word referring to the word list
 */
void complete_word()
{
    int i, j;
  
    if(pos) {
	/* find word start */
	for(j = pos; j > 0 && strchr(DELIM, edbuf[j - 1]) == NULL; j--);
	if(j == pos) return;
     
	if(edbuf[j] == '#' && pos - j > 1) {
	    /* try to complete built-in command */
	    char *p;
	    for(i = 0; (p = commands[i].name); i++) {
		if(strncmp(p, edbuf + j + 1, pos - j - 1) == 0)
		    break;
	    }
	    if(p)
		insertchars(p + pos - j - 1, strlen(p + pos - j) + 1);
	} else {
	    /* scan word list for match */
	    for(i = wordindex - 1; i != wordindex; i--) {
		if(i < 0) i += WORDSKEPT;
		if(words[i] == NULL) return;
		if(strncasecmp(words[i], edbuf + j, pos - j) == 0) break;
	    }
	    if(i != wordindex)
		insertchars(words[i] + pos - j, strlen(words[i] + pos - j));
	}
    }
}

/*
 * match and complete entire lines backwards in history
 */
void complete_line()
{
  int i;
  
  if(edlen) {
    for(i = curline - 1; i != curline; i--) 
     {
       if(i < 0) i = MAXHIST - 1;
       if(hist[i] && strncmp(edbuf, hist[i], edlen) == 0) break;
     }
    if(i != curline) 
     {
       clearinpline();
       strcpy(edbuf, hist[i]);
       pos = edlen = strlen(edbuf);
       redraw_input();
     }
  }
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
  char *p, *q, *r;
  char buf[LINELEN];
  int n;
  
  strcpy(buf, edbuf);
  p = buf; r = edbuf;
  while(*p) 
   {
     while(*p && !isalnum(*p)) *r++ = *p++;
     q = p;
     while(*p && isalnum(*p)) p++;
     if(*p || *q) 
      {
	n = p - q;
	if(strncasecmp(q, "u", n) == 0) 
	 {
	   strcpy(r, "you"); r += 3;
	 }
	else if(strncasecmp(q, "ur", n) == 0) 
	 {
	   strcpy(r, "your"); r += 4;
	 }
	else if(strncasecmp(q, "urs", n) == 0) 
	 {
	   strcpy(r, "yours"); r += 5;
	 }
	else if(strncasecmp(q, "r", n) == 0) 
	 {
	   strcpy(r, "are"); r += 3;
	 }
	else if(strncasecmp(q, "4", n) == 0) 
	 {
	   strcpy(r, "for"); r += 3;
	 }
	else if(strncasecmp(q, "2", n) == 0) 
	 {
	   strcpy(r, "to"); r += 2;
	 }
	else if(strncasecmp(q, "b", n) == 0) 
	 {
	   strcpy(r, "be"); r += 2;
	 } 
	else if(strncasecmp(q, "b4", n) == 0) 
	 {
	   strcpy(r, "before"); r += 6;
	 }
	else 
	 {
	   strcpy(r, q); r += n;
	 }
      }
   }
  *r = '\0';
  edlen = r - edbuf;
}
#endif

/* editor utility functions: */

/* insert one character first in a string */
void strins(s, c)
char *s; char c;
{
  char d;
  for(; *s; d = *s, *s++ = c, c = d);
  *s++ = c; *s = 0;
}

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan()
{
  printf("%s#cancan stopped dancing.\n", edattrend);
  reset_terminal();
  exit(0);
}
 
#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(arg)
char *arg;
{
    signal(SIGINT, SIG_IGN);
    if(!(waiting_for_pid = fork())) {
	static char *args[4] = { "/bin/sh", "-c", NULL, NULL };
	args[2] = arg;
	execvp(args[0], args);
	_exit(1);
    } else if(waiting_for_pid < 0) {
	printf("#fork failed, sorry\n");
    } else {
	while(waiting_for_pid) pause();
    }
    signal(SIGINT, intrsig);
    return wait_result;
}
#endif

/*
 * set up our signal handlers
 */
void setsigs()
{
  signal(SIGCHLD, childsig);
  signal(SIGINT, intrsig);
  signal(SIGTERM, termsig);
  signal(SIGPIPE, pipesig);
  signal(SIGWINCH, winchsig);
}

void pipesig()
{
  printf("\n#broken pipe.\n");
  exitcancan();
}

void intrsig()
{
  printf("\n#interrupted.\n");
  exitcancan();
}

void termsig()
{
  printf("\n#termination signal.\n");
  exitcancan();
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	signal(SIGTSTP, SIG_DFL);
	signal(SIGTERM, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	if(edattr) { edattr = 0;  fputs(edattrend, stdout); }
	reset_terminal();  fflush(stdout);
	if(kill(0, SIGTSTP) == -1) syserr("susp cancan");
	set_terminal();
	setsigs();
	winchsig();		/* in case size changed meanwhile */
	gotoxy(0, line0 = lines_1);	/* so we know where the cursor is */
	check_prompt();
	redraw_input();
    } else
	printf("\n#your shell probably lacks job control.\n");
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
  fprintf(stderr, "cancan system call error: %s (%d", msg, errno);
  if(errno > 0 && errno < sys_nerr)
     fprintf(stderr, ": %s", sys_errlist[errno]);
  fprintf(stderr, ")\n");
  reset_terminal();
  exit(1);
}


/*
 * redisplay the prompt if needed
 */
void check_prompt()
{
    if(data_sent)
	promptstr[0] = col0 = 0;
    else if(col0) {
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
	mark_and_print(promptstr);
    }
}

void def_error(ln)
int ln;
{
    static int n_err = 0;
    if(n_err++ >= 10) exit(1);
    printf("bad definition file format, line %d\n", ln);
}

/*
 * read definitions from file
 */
void read_settings(deff)
char *deff;
{
    char buf[BUFSIZE];		/* sizeof(buf) should be >= 2 * LINELEN */
    FILE *f = fopen(deff, "r");
    char *p, *cmd;
    int ln;

    if(!f) syserr("fopen");
    for(ln = 1; fgets(buf, sizeof(buf), f); ln++) {
	cmd = p = buf;
	while(*p != ' ' && *p) p++;
	if(*p && *cmd++ == '#') {
	    *p++ = 0;
	    strtok(p, "\n");	/* strip lf */
	    if(strcmp(cmd, "savefile") == 0) {
		if(atoi(p) > SAVEFILEVER)
		    printf("Warning: this cancan version is too old!\n");
	    } else if(strcmp(cmd, "alias") == 0) {
		parse_alias(p);
	    } else if(strcmp(cmd, "host") == 0) {
		if(!(p = strtok(p, " ")))
		    def_error(ln);
		else {
		    strcpy(hostname, p);
		    if((p = strtok(NULL, " ")))
			strcpy(portnumber, p);
		}
	    } else if(strcmp(cmd, "init") == 0) {
		strcpy(initstr, p);
	    } else if(strcmp(cmd, "action") == 0) {
		parse_action(p);
	    } else if(strcmp(cmd, "mark") == 0) {
		parse_marker(p);
	    } else if(strcmp(cmd, "bind") == 0) {
		parse_bind(p);
	    } else {
		def_error(ln);
	    }
	} else
	    def_error(ln);
    }
    fclose(f);
}

/*
 * save settings in definition file
 */
void save_settings()
{
    char buf[LINELEN], *cp;
    FILE *f;
    union {
	aliasnode *al;
	actionnode *ac;
	marknode *m;
	keynode *k;
    } p;
  
    if(!(cp = getenv("SAVEFILE"))) return;
    if(!(f = fopen(cp, "w"))) {
	printf("#error: can't write to definition file %s\n", cp);
	return;
    }
    fprintf(f, "#savefile %d\n", SAVEFILEVER);
    fprintf(f, "#host %s %s\n", hostname, portnumber);
    if(*initstr)
	fprintf(f, "#init %s\n", initstr);
    for(p.al = aliases; p.al; p.al = p.al->next)
	fprintf(f, "#alias %s=%s\n", p.al->name, p.al->subst);
    for(p.ac = actions; p.ac; p.ac = p.ac->next) {
	escape_specials(p.ac->pattern, buf, "\\\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (p.ac->active) ? '+' : '-',
		p.ac->label, buf, p.ac->commands);
    }
    for(p.m = markers; p.m; p.m = p.m->next) {
	escape_specials(p.m->pattern, buf, "\\=");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(p.m->attrcode));
    }
    for(p.k = keydefs; p.k; p.k = p.k->next)
	if(p.k->funct == key_run_command) {
	    escape_seq(p.k->sequence, buf);
	    fprintf(f, "#bind %s %s=%s\n", p.k->name, buf, p.k->call_data);
	}
    fclose(f);
}

/*
 * if alias recording is active, record given line in that alias
 */
void record_line(line)
char *line;			/* mustn't be trashed! */
{
    aliasnode **np;
    char new_line[LINELEN];
    if(strcmp(line, recording_name) == 0) return;
    if(!*(np = lookup_alias(recording_name))) {
	add_aliasnode(recording_name, line);
    } else {
	if(strlen((*np)->subst) + strlen(line) + 1 >= LINELEN - 16) {
	    /* too long, must be split */
	    char new_name[LINELEN];
	    int i = 0;		/* try name1, name2, ... */
	    do {
		sprintf(new_name, "%s%d", recording_name, ++i);
	    } while(*lookup_alias(new_name));
	    /* now we have a fresh new name */
	    add_aliasnode(new_name, (*np)->subst);
	    strcpy(new_line, new_name);
	} else
	    strcpy(new_line, (*np)->subst);
	strcat(new_line, ";");
	strcat(new_line, line);
	free((*np)->subst);
	(*np)->subst = strdup(new_line);
    }
}

#ifdef NOSTRDUP
/*
 * some old libraries don't have strdup(), so...share and enjoy. 
 */
char *strdup(s)
char *s;
{
  char *p = (char*)malloc(strlen(s) + 1);
  if(!p) syserr("malloc");
  return strcpy(p, s);
}
#endif

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions:
 *
 * note on con_sess[MAXFDSCAN]:
 *	con_sess[0].active holds total no. of connections
 *	con_sess[1].active holds main socket descriptor
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess ()
{
  return (sockfd == con_sess[1].active);
}

/*
 * return main session sockfd
 */
int main_sess ()
{
  return con_sess[1].active;
}

/*
 * send a string to the main session on the remote host
 */
void send_to_main_host(str, len)
char *str;
int len;
{
    int osockfd = sockfd;	/* backup current socket fd */
    sockfd = main_sess();
    send_to_host(str, len);
    sockfd = osockfd;
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if null or invalid id is given
 */
int find_sess (id)
char *id;
{
  int i, n;
  
  for (i=3, n=con_sess[0].active; n; i++)
     if (con_sess[i].id) 
      {
	n--;
	if (!strcmp(con_sess[i].id, id))
	   return i;
      }
  printf("#no such connection\n");
  return 0;
}

/*
 * connect another session
 */
void connect_sess (id, init)
char *id;	/* session id - NULL if main session */
char *init;	/* optional init string */
{
  int backfd = sockfd;
  /* dial the number by moving the right index in small circles */
  if ((sockfd = doconnect(hostname, atoi(portnumber))) < 0) 
   {
     if (id)
     	printf("host not responding or too many open connections\n");
     else	/* main session */
        syserr("#connect: host not responding");
   }
  else 
   {
     FD_SET(sockfd, &fdset);	 /* add socket to select() set */
     con_sess[0].active++;
     if (!id) 
      {  /* main session */
	con_sess[1].active = sockfd;  /* save main socket fd */
	con_sess[sockfd].active = 1;
	con_sess[sockfd].id = "main";
      } 
     else
      {
	if (!init || !*init)
	   con_sess[sockfd].active = 1;  /* active only if no init */
	con_sess[sockfd].id = (char *)malloc(strlen(id) + 1);
	if (!con_sess[sockfd].id) syserr("malloc");
	strcpy(con_sess[sockfd].id, id);
      }
     if (init)
     	parse(init, 0, NULL);
   }
   if (id)
      sockfd = backfd;
}

/*
 * disconnect session
 */
void disconnect_sess (id)
char *id;
{
  int sfd;
  
  if (id)
   {  /* #disconnect cmd */
     if (!(sfd = find_sess(id)))
     	return;
     close(sfd);
   } 
  else
     sfd = sockfd;  /* connection closed by remote host */
  
  printf("#connection \"%s\" closed.\n", con_sess[sfd].id);
  if (sfd == con_sess[1].active)  /* main connection closed */
    exitcancan();
  
  FD_CLR(sfd, &fdset);
  con_sess[0].active--;
  con_sess[sfd].active = 0;
  free(con_sess[sfd].id);
  con_sess[sfd].id = 0;
}

/*
 * toggle output display from another session
 */
void snoop_sess (id)
char *id;
{
    int sfd;
  
    sfd = find_sess(id);
    if(sfd)
	con_sess[sfd].active ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int i;
    int n = con_sess[0].active;
    printf("#The following connection%s open:\n", (n == 1) ? " is" : "s are");
    for(i = 3; n; i++)
	if(con_sess[i].id) {
	    n--;
	    printf("##%-8s %s\n", con_sess[i].id,
		con_sess[i].active ? (i == 3 ? "(active)" : "(snooped)") : "");
	}
}
#endif

#ifdef NOSETENV
int setenv(variable, value, overwrite)
char *variable, *value;
int overwrite;
{
    char buf[LINELEN];
    sprintf(buf, "%s=%s", variable, value);
    putenv(strdup(buf));
    return 1;
}

void unsetenv(variable)
char *variable;
{
    extern char **environ;
    int i, len = strlen(variable);
    for(i = 0; environ[i]; i++)
	if(!strncmp(environ[i], variable, len)) {
	    free(environ[i]);
	    while((environ[i] = environ[i + 1])) i++;
	}
}
#endif
