/*
 * cancan.h
 * various constants etc
 */

/* CONSTANTS */

#define CMDSEP ';'		/* command separator character */
#define DELIM " ;:.?!,'"	/* word delimiting characters */
#define MAXFDSCAN 32		/* max number of fd to scan in select() */
#define CAPLEN 20		/* max length of a terminal capability */
#define MAXARGS 16		/* max number of arguments to editor */
#define MAXSUBOPT 256		/* max length of suboption string */
#define BUFSIZE 2048		/* general i/o buffer size */
#define LINELEN 1024		/* maximum line length */
#define MAXHIST 128		/* number of history lines kept */
#define MAXMAPLEN 512		/* maximum length of automapped path */
#define WORDSKEPT 512		/* number of words kept for TAB-completion */
#define FLASHDELAY 500000	/* time of parentheses flash in microsecs */
#define KBD_TIMEOUT 35000	/* timeout for keyboard read in microsecs */

#define LM_NOECHO 1		/* no local echo */
#define LM_CHAR 2		/* char-by-char mode (no line editing) */

/* shared functions: */

/* from cancan.c */
void printver();
void flashparen();
void swap_history();
void put_in_history();
void put_word();
void complete_word();
void complete_line();
void strins();
void exitcancan();
int execute();
void setsigs();
void intrsig();
void suspend_cancan();
void syserr();
void check_prompt();
void save_settings();
void record_line();
int is_main_sess();
int main_sess();
void send_to_main_host();
int find_sess();
void connect_sess();
void disconnect_sess();
void snoop_sess();
void show_connections();

/* from parse.c */
void parse();
int search_action();
char *chop_arg();
void escape_command();
char *escape_specials();
void unescape();
char *skipspace();
char *skipnum();

/* from mapping.c */
void remove_steps();
void retrace_steps();
void show_automap();
void add_dir_to_map();

/* from tel.c */
int doconnect();
int read_socket();
void send_term_size();
void send_to_host();
void flush_output_buffer();

/* shared vars */

/* from cancan.c */
extern int mp_active;
extern int new_settings;	/* action/alias/etc has changed */
extern int verbose;		/* this is set if cmd is user-typed */
extern int wrap;
extern int speedwalk;
extern int verbatim;

extern int linemode;		/* line mode with flags as above (LM_*) */
extern int sockfd;		/* socket file descriptor */
extern int waiting_for_pid;
extern int wait_result;

extern int cols_1, lines_1;	/* for efficiency's sake */
extern int cols, lines;		/* terminal window size */
extern int line0, col0;		/* origin of input line */
extern char promptstr[];	/* prompt string */
extern char edbuf[];		/* input line buffer */
extern int edlen;		/* characters in edbuf */
extern int pos;			/* cursor position in edbuf */
extern int edattr;
extern char edattrbeg[], edattrend[];

extern char initstr[];

extern FILE *capfile;
extern char prefixstr[];
extern char inserted_next[];
extern char footnote[];

extern char *hist[];
extern int curline;
extern int pickline;

extern char *recording_name;

/* from tel.c */
extern int data_sent;		/* incremented each time we send_to_host */

/* ugly debug macro hack, vt100-like only */
#define NOTE(param) (printf("\0337\033[H\033[K"),printf param,\
		     printf("\0338"),fflush(stdout))

/* ASCII assumed */
#define ctoi(c) ((int)((unsigned char)(c) - '0'))
#define CONTROL(c) ((unsigned char)(c) & 0x1f)

/* macros to find cursor position from input buffer position */
#define CURLINE(pos) (((pos) + col0) / cols_1 + line0)
#define CURCOL(pos) (((pos) + col0) % cols_1)

/* macros to match parentheses */
#define ISRPAREN(c) ((c) == ')' || (c) == ']' || (c) == '}')
#define ISLPAREN(c) ((c) == '(' || (c) == '[' || (c) == '{')
#define LPAREN(c) ((c) == ')' ? '(' : ((c) == ']' ? '[' : '{'))

/*
 * Attribute codes: bit 0-3 for foreground colour or effects,
 * 4-7 for background colour. Colour #8 is "none", so 0x88 is "no attribute".
 */
#define COLOURS 8		/* number of ANSI colours */
#define NOATTRIB COLOURS	/* no colour change, use default */
#define BOLD (COLOURS + 1)
#define UNDERLINE (COLOURS + 2)
#define INVERSE (COLOURS + 3)
#define TOK_ON (COLOURS + 4)	/* these two used by attr_token() */
#define TOK_INVALID (-1)

/* constructors / accessors for attribute codes */
#define ATTRCODE(fg, bg) ((fg) | ((bg << 4)))
#define FOREGROUND(attrcode) ((attrcode) & 0xf)
#define BACKGROUND(attrcode) ((attrcode) >> 4)

#define NOATTRCODE ATTRCODE(NOATTRIB, NOATTRIB)

/*
 * NCSA telnet 2.2 doesn't reset the colour when it receives "esc [ m",
 * so we must know what the normal colours are in order to reset it.
 * These colours can be changed with the #stdcolour command.
 */
#ifdef TELNETBUG
#define DEFAULTFG 7		/* make default white text */
#define DEFAULTBG 4		/* on blue background */
#endif

#ifdef TERM
#define OMIT_SESSIONS
#endif

#ifdef OMIT_SESSIONS
#define send_to_main_host(buf, len) (send_to_host(buf, len))
#define main_sess() (sockfd)
#define is_main_sess() (1)
#endif
