/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "cancan.h"
#include "edit.h"
#include "term.h"

char *strdup();

/*
 * insert a typed character (if it is printable)
 */
void insert_char(c)
int c;
{
    static char ch;
    if(edlen < LINELEN - 1 && c >= ' ') {
	if(!edattr && *edattrbeg) {
	    edattr = 1;  fputs(edattrbeg, stdout);
	}
	ch = c;
	insertchars(&ch, 1);
	if(ISRPAREN(c)) flashparen(c);
    }
}

/*
 * redraw the input line
 */
void redraw_input_line()
{
    clearinpline();
    redraw_input();
}

/*
 * transpose two characters to the left
 */
void transpose_chars()
{
    int i, cl, cc;
    unsigned char c;
    if(pos > 1) {
	cl = CURLINE(pos); cc = CURCOL(pos);
	if(pos < edlen)
	    pos++;
	i = pos - 2;
	c = edbuf[i + 1]; edbuf[i + 1] = edbuf[i]; edbuf[i] = c;
	optmove(cl, cc, CURLINE(i), CURCOL(i));
	putchar(edbuf[i]);
	optmove(CURLINE(i), CURCOL(i) + 1, CURLINE(i + 1), CURCOL(i + 1));
	putchar(edbuf[i + 1]);
	optmove(CURLINE(i + 1), CURCOL(i + 1) + 1, CURLINE(pos),
		CURCOL(pos));
    }
}

/*
 * erase everything to the end of line
 */
void kill_to_eol()
{
    int cline = CURLINE(pos);
    fputs(clreoln, stdout);
    if(CURLINE(edlen) > cline) {
	printf("\n%s", clreoscr);
	optmove(cline + 1, 0, cline, CURCOL(pos));
    }
    edbuf[pos] = '\0';
    edlen = pos;
}

/*
 * move cursor to end of line
 */
void end_of_line()
{
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(edlen), CURCOL(edlen));
    pos = edlen;
}

/*
 * move cursor to beginning of line
 */
void beginning_of_line()
{
    optmove(CURLINE(pos), CURCOL(pos), line0, col0);
    pos = 0;
}

/*
 * delete a character to the right
 */
void del_char_right()
{
    if(pos < edlen)
	deletechars(1);
}

/*
 * delete a character to the left
 */
void del_char_left()
{
    if(pos) {
	pos--;
	optmove(CURLINE(pos + 1), CURCOL(pos + 1), CURLINE(pos), CURCOL(pos));
	deletechars(1);
    }
}

/*
 * enter a line into history, but don't do anything else
 */
void enter_into_history()
{
    clearinpline();
    put_in_history(edbuf);
    pickline = curline;
    strcpy(edbuf, prefixstr);
    pos = edlen = strlen(edbuf);
}

/*
 * enter a line
 */
void enter_line()
{
    char *p;
    int ls = (pos >= edlen) ? 1 : 1 + CURLINE(edlen) - CURLINE(pos);
    if((line0 += ls) > lines_1) line0 = lines_1;
    do printf("\n"); while(--ls); 
    if(capfile) fprintf(capfile, "%s\n", edbuf);
#ifdef NEWBIE
    denewbiefy();
#endif
    if(verbatim && edbuf[0] != '#') {
	edbuf[edlen] = '\n';
	send_to_host(edbuf, edlen + 1);
	edbuf[edlen] = '\0';
    } else {
	if(edattr) { edattr = 0; fputs(edattrend, stdout); }
	new_settings = 0;
        verbose = 1;
	parse(edbuf, 1, NULL);
	verbose = 0;
	if(new_settings)	/* user defined something -- save */
	    save_settings();
    }
    check_prompt();
    /* don't put identical lines in history, nor short ones */
    p = hist[curline ? curline - 1 : MAXHIST - 1];
    if(!p || (strcmp(edbuf, p) != 0 && edlen > 1))
	put_in_history(edbuf);
    pickline = curline;
    if(inserted_next[0]) {
	strcpy(edbuf, inserted_next);
	inserted_next[0] = '\0';
    } else
	strcpy(edbuf, prefixstr);
    pos = edlen = strlen(edbuf);
    redraw_input();
}

/*
 * move one word forward
 */
void next_word()
{
    int i;
    for(i = pos; !isalnum(edbuf[i]) && edbuf[i]; i++);
    while(isalnum(edbuf[i])) i++;
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(i), CURCOL(i));
    pos = i;
}

/*
 * move one word backward
 */
void prev_word()
{
    int i;
    for(i = pos; i && !isalnum(edbuf[i - 1]); i--);
    while(i && isalnum(edbuf[i - 1])) i--;
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(i), CURCOL(i));
    pos = i;
}

/*
 * delete word to the right
 */
void del_word_right()
{
    int i;
    for(i = pos; !isalnum(edbuf[i]) && i < edlen; i++);
    while(isalnum(edbuf[i])) i++;
    deletechars(i - pos);
}

/*
 * delete word to the left
 */
void del_word_left()
{
    int i;
    for(i = pos; i && !isalnum(edbuf[i - 1]); i--);
    while(i && isalnum(edbuf[i - 1])) i--;
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(i), CURCOL(i));
    i = pos - i; pos -= i;
    deletechars(i);
}

/*
 * get previous line from history list
 */
void prev_line()
{
    int i = pickline - 1;
    if(i < 0) i = MAXHIST - 1;
    swap_history(i);
}

/*
 * get next line from history list
 */
void next_line()
{
    int i = pickline + 1;
    if(i == MAXHIST) i = 0;
    swap_history(i);
}

/*
 * move one char backward
 */
void prev_char()
{
    if(pos) {
	pos--;
	optmove(CURLINE(pos + 1), CURCOL(pos + 1), CURLINE(pos), CURCOL(pos));
    }
}

/*
 * move one char forward
 */
void next_char()
{
    if(pos < edlen) {
	pos++;
	optmove(CURLINE(pos - 1), CURCOL(pos - 1), CURLINE(pos), CURCOL(pos));
    }
}

/*
 * execute string as if typed
 */
void key_run_command(cmd)
char *cmd;
{
    clearinpline();
    if(!edattr) fputs(edattrbeg, stdout);
    printf("%s%s\n", cmd, edattrend);  edattr = 0;
    if(capfile) fprintf(capfile, "%s\n", cmd);
    parse(cmd, 1, NULL);
    check_prompt();
    redraw_input();
}
