/* linked list operations */

/* generic linked list node */
typedef struct defnode {
    struct defnode *next;
    char *field;
} defnode;

/* linked list nodes: keep the next pointer first, then string to sort by */
typedef struct aliasnode {
    struct aliasnode *next;
    char *name;
    char *subst;
    int active;
} aliasnode;

typedef struct actionnode {
    struct actionnode *next;
    char *pattern;
    char *commands, *label;
    int active;
} actionnode;

typedef struct marknode {
    struct marknode *next;
    char *pattern, *pos;
    int attrcode;
} marknode;

typedef struct keynode {
    struct keynode *next;	/* next key definition in list */
    char *sequence;		/* escape sequence sent by terminal */
    char *name;			/* key name */
    void (*funct)();		/* function called when key pressed */
    char *call_data;		/* data passed to function */
} keynode;

extern aliasnode *aliases;
extern actionnode *actions;
extern marknode *markers;
extern keynode *keydefs;

void add_node();
void add_aliasnode();
void add_actionnode();
void add_marknode();
void add_keynode();

defnode **lookup();
aliasnode **lookup_alias();
actionnode **lookup_action();
actionnode **lookup_action_pattern();
marknode **lookup_marker();
keynode **lookup_key();

void delete_aliasnode();
void delete_actionnode();
void delete_marknode();
void delete_keynode();

/* generic lookup used instead for searching by node's first data field */
#define lookup_alias(name) ((aliasnode **)lookup(name, &aliases))
#define lookup_action_pattern(pattern) ((actionnode **)lookup(pattern, &actions))
#define lookup_marker(pattern) ((marknode **)lookup(pattern, &markers))
