/*
 * parse.c
 * command parsing, argument parsing routines
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "list.h"

void bzero();
#ifdef NOBZERO
#define bzero(a,b) memset ((a), 0, (b))
#endif

struct param {
    int len;
    char *s;
};

static void subst_param();
static int parse_walk();
static int match_action();
static void split_arg();
static char *skip_arg();


/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params may be given.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static defnode *recursed_aliases;
    aliasnode *np;
    char cmd[LINELEN], buf[LINELEN], *s, *d;
    int i, n_cmd = 1;
    int subst = 0;
    if(*(s = skipspace(cmdline)) == '{') {	/* check if enclosed in {} */
	if(!(d = skip_arg(s, 0))) return;
	if(!*skipspace(d))
	    s++, d--, strncpy(cmd, s, d - s), cmd[d - s] = 0, subst = 1;
    }
    if(!subst) strcpy(cmd, cmdline);

    for(s = cmd; ; *s++ = 0, n_cmd++)		/* chop into commands */
	if(!(s = skip_arg(s, CMDSEP))) return;
	else if(!*s) break;
    
    /* PROCESS THE COMMANDS ONE BY ONE */
    for(s = cmd; n_cmd--; s = cmdline) {
	cmdline = s + strlen(s) + 1;
	if(recording_name && !param && *s != '#') record_line(s);

	if(*(d = skip_arg(s, ' '))) *d = 0;	/* pick out the first word */
	else d = 0;

	if((np = *lookup_alias(s)) && *lookup(s, &recursed_aliases)) np = 0;

	if(np) { if(d) *d = ' '; }
	else if(*s == '#') { if(d) d = skipspace(d + 1); }
	else { if(d) *d = ' '; d = s; }		/* restore line */

	buf[0] = buf[sizeof(buf) - 1] = 0;
	if(d)
	    if(subst) subst_param(buf, d, param);
	    else strcpy(buf, d);
	if(param && !n_cmd && !subst && recursed_aliases && param[0].len) {
	    i = strlen(buf);		/* append arguments */
	    strncpy(buf + i, param[0].s, sizeof(buf) - i);
	}
	if(buf[sizeof(buf) - 1]) {
	    buf[sizeof(buf) - 1] = 0;
	    printf("#warning: line truncated\n");
	}

	if(np) {				/* process an alias */
	    struct param new_param[10];
	    defnode ban_alias;
	    ban_alias.next = recursed_aliases;  ban_alias.field = s;
	    split_arg(buf, new_param);
	    recursed_aliases = &ban_alias;	/* block alias recursion */
	    parse(np->subst, silent, new_param);
	    recursed_aliases = ban_alias.next;
	} else if(*s != '#') {			/* text to send		*/
	    if(parse_walk(buf)) continue;
	    if(!subst) unescape(buf);
	    if(!silent) printf("[%s%s%s]\n", edattrbeg, buf, edattrend);
	    buf[i = strlen(buf)] = '\n';
	    send_to_host(buf, i + 1);
	} else if(*++s) {			/* "# " is a comment	*/
	    if(isdigit(*s) && !*skipnum(s)) {
		if((i = atoi(s)) > 0)
		    while(i--) parse(buf, 1, NULL);
		else 
		    printf("#bogus repeat count\n");
	    } else
#ifndef OMIT_SESSIONS
	    if(*s == '#') {			/* a session is specified */
		int backfd = sockfd;
		int back_data_sent = data_sent;
		if((sockfd = find_sess(s + 1))) {
		    parse(buf, 1, param);	/* do this always silently */
		    if(!is_main_sess())
			data_sent = back_data_sent;
		}
		sockfd = backfd;
	    } else if(!is_main_sess())
		printf("#command not allowed on subsidiary connection\n");
	    else
#endif
#ifndef NOSHELL
	    if(*s == '?' && !*(s + 1)) {	/* #? command	*/
		if(!n_cmd) {
		    execute(buf);
		} else if(execute(buf)) {
		    cmdline += strlen(cmdline) + 1;  --n_cmd;
		} else {
		    parse(cmdline, silent, param);  return;
		}
	    } else
#endif
	    if(param && !recursed_aliases && !strcmp(s, "add")) {
		if(param[0].len)
		    sprintf(buf, "%.*s", param[0].len, param[0].s);		
		if(*buf) put_word(buf);
	    } else
	    for(i = 0; ; i++)
		if(!commands[i].name) {
		    printf("#unknown cancan command '%s'\n", s);
		    break;
		} else if(commands[i].funct && !strcmp(s, commands[i].name)) {
		    (*commands[i].funct)(buf);
		    if(line0 < lines_1)
			gotoxy(0, line0 = lines_1);
		    break;
		}
	}
    }	/* END OF THE LOOP */
}

/*
 * {} *must* be balanced in input string. input will be thrashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &(command)... return 1 on success.
 */
static void subst_param(buf, str, param)
char buf[LINELEN], *str;
struct param param[10];
{
    static char specials[] = "&\\{";
    char *src, *s = str;
    int len, room = LINELEN;

    while(*s) {
	if(*s == '\\') {
	    if(!*++s) break;
	    src = s;  s += len = strcspn(s + 1, specials) + 1;
	} else if(*s == '{') {
	    src = s;  len = (s = skip_arg(s, 0)) - src;
	} else if(*s != '&') {
	    src = s;  s += len = strcspn(s, specials);
	} else switch(*++s) {
	case '&':				/* delayed &	*/
	    src = s;  for(len = 1; *++s == '&'; len++);
	    break;
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
	    *s++ = 0;  unescape(src);
	    if(!(src = getenv(src))) continue;
	    len = strlen(src);
	    break;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(*s) {
		FILE *fp;
		*s++ = 0;
		if((fp = popen(src, "r"))) {
		    if((len = fread(buf, 1, room, fp)) > 0)
			buf += len, room -= len;
		    pclose(fp);
		}
	    } else
		printf("#unbalanced ( near \"&(%.16s\"\n", src);
	    continue;
	default:				/* &N or &label	*/
	    if(isdigit(*s)) {
		len = ctoi(*s++);
		if(!param || !param[len].len) continue;
		src = param[len].s;  len = param[len].len;
	    } else {
		for(src = s; isalpha(*s); s++);
		len = *s;  *s = 0;
		src = getenv(src);  *s = len;
		if(!src) continue;
		len = strlen(src);
	    }
	} /* switch ends */
	if(room < len) len = room;
	strncpy(buf, src, len);
	buf += len;  room -= len;
    }
    if(room) *buf = 0;
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
static int parse_walk (word)
char *word;
{
    long strtol();
    char buf[16];
    int n = strlen(word);
  
    if(!n || (n > 1 && !speedwalk)
	|| !strchr("neswud", word[n - 1])
	|| strspn(word, "neswud0123456789") != n)
	return 0;

    strcpy(buf, " \n");
    while (*word) {
	if(isdigit(*word))
	    n = strtol(word, &word, 10);
	else
	    n = 1;
	while(n--) {
	    buf[0] = *word;
	    send_to_host(buf, 2);
	    if(is_main_sess())
		add_dir_to_map(*word);
	}
	word++;
    }
    return 1;
}

/*
 * match action containing &0..&9 and return actual params in param array -
 * return 1 if matched, 0 if not
 */
static int match_action (pat, src, param)
char *pat, *src;
struct param param[10];
{
  char mpat[LINELEN], *npat=0, *nsrc=0;
  int mbeg = 0, mword = 0, prm = -1, p;
  
  if (*pat == '^') 
   {
     pat++;
     mbeg = 1;  /* anchor match at line start */
   }
  if (*pat == '&') mbeg = - mbeg - 1;  /* pattern starts with '&' */
  
  while (pat && *pat) 
   {
     if (*pat == '&') 
      {
	pat++;
	if (isdigit(*pat)) 
	 {
	   prm = ctoi(*pat);
	   pat++;
	 }
	else 
	 {
	   prm = 0;  /* support single & for back compat */
	   mword = 1;	 /* match a single word only */
	 }
      }
     
     npat = strchr(pat, '&');
     if (npat) 
      {
	strncpy(mpat, pat, npat-pat);
	mpat[npat - pat] = 0;
      }
     else
     	strcpy(mpat, pat);
     
     if (*mpat) 
      {
	nsrc = strstr(src, mpat);
	if (!nsrc) return 0;
	if (mbeg > 0) 
	 {
	   if (nsrc != src) return 0;
	   mbeg = 0;  /* reset mbeg to stop further start match */
	 }
	if (prm != -1)
	 {
	   param[prm].len = nsrc-src;
	   param[prm].s = src;
	 }
	}
     else if (prm != -1) 
      {
	/* end of pattern space, strip trailing blanks */
	for (p = strlen(src) - 1; p>=0 && src[p]==' '; p--);
	param[prm].len = p + 1;
	param[prm].s = src;
      }
     
     /* post-processing of param */
     if (prm != -1) 
      {
	if (mword) 
	 {
	   if (mbeg == -1) 
	    {
	      /* unanchored '&' start, take last word */
	      p = param[prm].len;  src = param[prm].s;
	      while(p && src[p - 1] == ' ') p--;
	      param[prm].len = p;
	      while(p && src[p - 1] != ' ') p--;
	      if(p)
	       {
	         param[prm].len -= p;
	         param[prm].s = src + p;
	       }
	    } 
	   else if ((src = strchr(param[prm].s, ' ')))
	    {
	      if(param[prm].len > (p = src - param[prm].s))
	         if(pat[1])	/* match only if param is single-worded */
	            return 0;
	         else		/* '&' at end of pattern, take first word */
	            param[prm].len = p;
	    }
	   mbeg = mword = 0;  /* reset match flags */
	    }
      }
     
     src = nsrc + strlen(mpat);
     pat = npat;
   }
  return 1;
}

/*
 * Search for actions to trig on an input line. The line can't be trashed
 * since we want to print it on the screen later.
 * Return 1 if line should be printed, 0 if it should be suppressed.
 */
int search_action(line)
char *line;
{
    actionnode *p;
    struct param param[10];
    bzero((char *)param, sizeof(param));
    while(*line == '\r') line++;
    for(p = actions; p; p = p->next)
	if(p->active && match_action(p->pattern, line, param)) {
	    if(strcmp("#gag", p->commands) == 0) return 0;
	    if(col0) printf("%s%s", begoln, clreoln);
	    parse(p->commands, 0, param);
	    if(col0) check_prompt();	/* prompt is redisplayed there */
	}
    return 1;
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_arg(str, param)
char *str;
struct param param[10];
{
    char *s = str;
    int i;
    param[0].s = str;
    param[0].len = strlen(str);
    for(i = 1; *s && i < 10; i++) {
	while(*s == ' ') s++;
	param[i].s = s;
	while(*s && *s != ' ' && (*s != '\\' || *++s)) s++;
	param[i].len = s - param[i].s;
    }
    while(i < 10) param[i++].len = 0;
}

/*
 * convert str ([-]NNN..) to int, return pointer to rest, or NULL.
 */
/*
char *get_num(num, str)
int *num; char *str;
{
    int val = 0, sign = 0;
    unsigned i;
    while(*str == ' ') str++;
    if(*str == '-') str++, sign = 1;
    do {
	if((i = ctoi(*str)) < 10)
	    val = 10 * val + i;
	else
	    return NULL;
    } while(*++str && *str != ' ');
    *num = sign ? -val : val;
    return str;
}
*/

/*
 * skip over a delim-terminated argument, which may contain (nested) {}.
 * if delim is 0, skip over one char or \char or {...} only.
 * return tail on success, NULL when {} are unbalanced.
 */
static char *skip_arg(str, delim)
char *str;
char delim;
{
    char *s = str;
    int nested = 0;
    if(*s && *s != delim) do
	if(*s == '\\') { if(!*++s) break; }
	else if(*s == '{') { str = s; nested++; }
	else if(*s == '}') { if(!nested--) break; }
    while(*++s && (nested || (delim && delim != *s)));
    if(!nested) return s;
    printf("#unbalanced %c near \"%.16s\"\n", (nested > 0) ? '{' : '}',
		(nested > 0 || s - str < 16) ? str : s - 15);
    return 0;
}

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = str;
    int nested = 0;
    if(*s && *s != delim) do
	if(*s == '\\') { if(!*++s) return 0; }
	else if(*s == '{') nested++;
	else if(*s == '}') nested--;
    while(*++s && (*s != delim || nested ));
    return (*s) ? *s = 0, s + 1 : 0;
}

/*
 * When needed, enclose in {} and escape &\; with a backslash.
 */
void escape_command(str, buf)
char str[LINELEN], buf[LINELEN];
{
    char *d = buf, *s = skip_arg(str, CMDSEP);
    if(s && *s) {
	for(*d++ = '{', s = str; d - buf < LINELEN - 1; *d++ = *s++)
	    if(*s == CMDSEP || *s == '&' || *s == '\\') *d++ = '\\';
	    else if(!*s) {
		*d++ = '}';  *d = 0;  break;
	    }
    } else
	strcpy(d, str);
    buf[LINELEN - 1] = 0;
}

/*
 * escape special chars with a backslash
 */
char *escape_specials(str, buf, spec)
char *str, *buf, *spec;
{
    int k;
    char *s = str, *d = buf;
    for(s = str, d = buf; ; *d++ = '\\', *d++ = *s++) {
	k = strcspn(s, spec);
	strncpy(d, s, k);
	d += k;  s += k;
	if(!*s) break;
    }
    *d = 0;
    return buf;
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *p;
    for(p = s; *s && (*s != '\\' || *++s); *p++ = *s++);
    *p = 0;
}

/*
 * return pointer to next non-blank char
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

/*
 * return pointer to next non-digit
 */
char *skipnum(p)
char *p;
{
    if(*p == '-') p++;
    while(isdigit(*p)) p++;
    return p;
}
